# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os

from abc import ABC, abstractmethod
from typing import Optional, Union, Tuple, Type

from nn4k.executor import NNExecutor
from nn4k.invoker.base import NNInvoker
from nn4k.utils.class_importing import dynamic_import_class


class NNHub(ABC):
    _hub_instance = None

    @staticmethod
    def get_instance() -> "NNHub":
        """
        Get the NNHub instance. If the instance is not initialized, create a stub `SimpleNNHub`.
        """
        if NNHub._hub_instance is None:
            NNHub._hub_instance = SimpleNNHub()
        return NNHub._hub_instance

    @abstractmethod
    def publish(
        self,
        model_executor: Union[NNExecutor, Tuple[Type[NNExecutor], tuple, dict, tuple]],
        name: str,
        version: str = None,
    ) -> str:
        """
        Publish a model(executor) to hub.

        :param model_executor: An NNExecutor object, which is pickleable.
                               Or a tuple of (class, init_args, kwargs, weight_ids) for creating an NNExecutor,
                               while all these 4 augments are pickleable.

        :param str name: The name of a model, like `llama2`. We do not have a `namespace`.
                         Use a joined name like `alibaba/qwen` to support such features.

        :param str version: Optional. Auto generate a version if this param is not given.

        :return: The published model version.
        :rtype: str
        """
        pass

    @abstractmethod
    def get_model_executor(
        self, name: str, version: str = None
    ) -> Optional[NNExecutor]:
        """
        Get an NNExecutor instance from Hub.

        :param str name: The name of a model.
        :param str version: The version of a model. Get default version of a model if this param is not given.
        :return: The ModelExecutor Instance. None for NotFound.
        :rtype: Optional[NNExecutor]
        """
        pass

    @abstractmethod
    def get_invoker(self, nn_config: dict) -> Optional["NNInvoker"]:
        """
        Get an NNExecutor instance from Hub.

        :param dict nn_config: The config dictionary.
        :return: The NNExecutor Instance. None for NotFound.
        :rtype: Optional[NNInvoker]
        """
        pass

    def start_service(self, name: str, version: str, service_id: str = None, **kwargs):
        raise NotImplementedError("This Hub does not support starting model service.")

    def stop_service(self, name: str, version: str, service_id: str = None, **kwargs):
        raise NotImplementedError("This Hub does not support stopping model service.")

    def get_service(self, name: str, version: str, service_id: str = None):
        raise NotImplementedError("This Hub does not support model services.")


class SimpleNNHub(NNHub):
    def __init__(self) -> None:
        super().__init__()
        self._model_executors = {}

    def _add_executor(
        self,
        executor: Union[NNExecutor, Tuple[Type[NNExecutor], tuple, dict, tuple]],
        name: str,
        version: str = None,
    ):
        from nn4k.consts import NN_VERSION_DEFAULT

        if version is None:
            version = NN_VERSION_DEFAULT
        if self._model_executors.get(name) is None:
            self._model_executors[name] = {version: executor}
        else:
            self._model_executors[name][version] = executor

    def publish(
        self, model_executor: NNExecutor, name: str, version: str = None
    ) -> str:
        from nn4k.consts import NN_VERSION_DEFAULT

        print(
            "WARNING: You are using SimpleNNHub which can only maintain models in memory without data persistence!"
        )
        if version is None:
            version = NN_VERSION_DEFAULT
        self._add_executor(model_executor, name, version)
        return version

    def _create_model_executor(self, cls, init_args, kwargs, weights):
        raise NotImplementedError()

    def get_model_executor(
        self, name: str, version: str = None
    ) -> Optional[NNExecutor]:
        from nn4k.consts import NN_VERSION_DEFAULT

        if version is None:
            version = NN_VERSION_DEFAULT
        if self._model_executors.get(name) is None:
            return None
        executor = self._model_executors.get(name).get(version)
        if isinstance(executor, NNExecutor):
            return executor
        cls, init_args, kwargs, weights = executor
        executor = self._create_model_executor(cls, init_args, kwargs, weights)
        return executor

    def _get_local_executor_class(self, nn_config: dict) -> Type[NNExecutor]:
        from nn4k.consts import NN_EXECUTOR_KEY, NN_EXECUTOR_TEXT
        from nn4k.consts import NN_NAME_KEY, NN_NAME_TEXT
        from nn4k.consts import NN_VERSION_KEY, NN_VERSION_TEXT
        from nn4k.consts import NN_LOCAL_HF_MODEL_CONFIG_FILE
        from nn4k.consts import NN_LOCAL_SENTENCE_TRANSFORMERS_CONFIG_FILE
        from nn4k.executor.huggingface.hf_embedding_executor import HFEmbeddingExecutor
        from nn4k.executor.huggingface.base.hf_llm_executor import HFLLMExecutor
        from nn4k.utils.config_parsing import get_string_field

        nn_executor = nn_config.get(NN_EXECUTOR_KEY)
        if nn_executor is not None:
            nn_executor = get_string_field(
                self.init_args, NN_EXECUTOR_KEY, NN_EXECUTOR_TEXT
            )
            executor_class = dynamic_import_class(nn_executor, NN_EXECUTOR_TEXT)
            if not issubclass(executor_class, NNExecutor):
                message = "%r is not an %s class" % (nn_executor, NN_EXECUTOR_TEXT)
                raise RuntimeError(message)
            return executor_class

        nn_name = nn_config.get(NN_NAME_KEY)
        if nn_name is not None:
            nn_name = get_string_field(nn_config, NN_NAME_KEY, NN_NAME_TEXT)
            if os.path.isdir(nn_name):
                file_path = os.path.join(nn_name, NN_LOCAL_HF_MODEL_CONFIG_FILE)
                if os.path.isfile(file_path):
                    file_path = os.path.join(
                        nn_name, NN_LOCAL_SENTENCE_TRANSFORMERS_CONFIG_FILE
                    )
                    if os.path.isfile(file_path):
                        executor_class = HFEmbeddingExecutor
                    else:
                        executor_class = HFLLMExecutor
                return executor_class

        nn_version = nn_config.get(NN_VERSION_KEY)
        if nn_version is not None:
            nn_version = get_string_field(nn_config, NN_VERSION_KEY, NN_VERSION_TEXT)
        message = "can not determine local executor class for NN config"
        if nn_name is not None:
            message += "; model: %r" % nn_name
            if nn_version is not None:
                message += ", version: %r" % nn_version
        raise RuntimeError(message)

    def get_invoker(self, nn_config: dict) -> Optional["NNInvoker"]:
        from nn4k.invoker import LLMInvoker
        from nn4k.invoker.openai import OpenAIInvoker
        from nn4k.utils.invoker_checking import is_openai_invoker, is_hub_invoker

        if is_openai_invoker(nn_config):
            invoker = OpenAIInvoker.from_config(nn_config)
            return invoker
        # TODO NN4K: this will be replaced once we publish the SimpleHub solution. Now we only have openai invoker,
        #  LLMInvoker and HubInvoker.
        elif is_hub_invoker(nn_config):
            from nn4k.invoker.hub_invoker import HubInvoker

            invoker = HubInvoker.from_config(nn_config)
            return invoker
        else:
            invoker = LLMInvoker.from_config(nn_config)
            return invoker



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
from nn4k.executor import LLMExecutor


class HFEmbeddingExecutor(LLMExecutor):
    @classmethod
    def from_config(cls, nn_config: dict) -> "HFEmbeddingExecutor":
        """
        Create an HFEmbeddingExecutor instance from `nn_config`.
        """
        executor = cls(nn_config)
        return executor

    def load_model(self, args=None, **kwargs):
        import torch
        from sentence_transformers import SentenceTransformer
        from nn4k.consts import NN_NAME_KEY, NN_NAME_TEXT
        from nn4k.consts import NN_VERSION_KEY, NN_VERSION_TEXT
        from nn4k.consts import NN_DEVICE_KEY
        from nn4k.utils.config_parsing import get_string_field

        nn_config: dict = args or self.init_args
        if self._model is None:
            nn_name = get_string_field(nn_config, NN_NAME_KEY, NN_NAME_TEXT)
            nn_version = nn_config.get(NN_VERSION_KEY)
            if nn_version is not None:
                nn_version = get_string_field(
                    nn_config, NN_VERSION_KEY, NN_VERSION_TEXT
                )
            model_path = nn_name
            revision = nn_version
            use_fast_tokenizer = False
            device = nn_config.get(NN_DEVICE_KEY)
            if device is None:
                device = "cuda" if torch.cuda.is_available() else "cpu"
            #
            # SentenceTransformer will support `revision` soon. See:
            #
            #   https://github.com/UKPLab/sentence-transformers/pull/2419
            #
            model = SentenceTransformer(
                model_path,
                device=device,
            )
            self._model = model

    def inference(self, inputs, **kwargs):
        model = self.model
        embeddings = model.encode(inputs)
        return embeddings



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional, Union


class NNExecutor(ABC):
    """
    Entry point of model execution in a certain pod.
    """

    def __init__(self, init_args: dict, **kwargs):
        self._init_args = init_args
        self._kwargs = kwargs
        self._model = None
        self._tokenizer = None

    @property
    def init_args(self):
        """
        Return the `init_args` passed to the executor constructor.
        """
        return self._init_args

    @property
    def kwargs(self):
        """
        Return the `kwargs` passed to the executor constructor.
        """
        return self._kwargs

    @property
    def model(self):
        """
        Return the model object managed by this executor.

        :raises RuntimeError: if the model is not initialized yet
        """
        if self._model is None:
            message = "model is not initialized yet"
            raise RuntimeError(message)
        return self._model

    @property
    def tokenizer(self):
        """
        Return the tokenizer object managed by this executor.

        :raises RuntimeError: if the tokenizer is not initialized yet
        """
        if self._tokenizer is None:
            message = "tokenizer is not initialized yet"
            raise RuntimeError(message)
        return self._tokenizer

    def execute_inference(self, args=None, **kwargs):
        """
        The entry point of batch inference in a certain pod.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support batch inference."
        )

    def inference(self, inputs, **kwargs):
        """
        The entry point of inference. Usually for local invokers or model services.
        """
        raise NotImplementedError()

    @abstractmethod
    def load_model(self, args=None, mode=None, **kwargs):
        """
        Implement model loading logic in derived executor classes.

        Implementer should initialize `self._model` and `self._tokenizer`.

        This method will be called by several entry methods in executors and invokers.
        """
        raise NotImplementedError()

    def warmup_inference(self, args=None, **kwargs):
        """
        Implement model warming up logic for inference in derived executor classes.
        """
        pass

    @classmethod
    @abstractmethod
    def from_config(cls, nn_config: Union[str, dict]) -> "NNExecutor":
        """
        Create an NN executor instance from `nn_config`.

        This method is abstract, derived class must override it by either
        creating executor instances or implementating dispatch logic.

        :param nn_config: config to use, can be dictionary or path to a JSON file
        :type nn_config: str or dict
        :rtype: NNExecutor
        """
        from nn4k.nnhub import NNHub
        from nn4k.consts import NN_NAME_KEY, NN_NAME_TEXT
        from nn4k.consts import NN_VERSION_KEY, NN_VERSION_TEXT
        from nn4k.consts import NN_EXECUTOR_KEY, NN_EXECUTOR_TEXT
        from nn4k.utils.config_parsing import preprocess_config
        from nn4k.utils.config_parsing import get_string_field
        from nn4k.utils.class_importing import dynamic_import_class

        nn_config = preprocess_config(nn_config)
        nn_executor = nn_config.get(NN_EXECUTOR_KEY)
        if nn_executor is not None:
            nn_executor = get_string_field(nn_config, NN_EXECUTOR_KEY, NN_EXECUTOR_TEXT)
            executor_class = dynamic_import_class(nn_executor, NN_EXECUTOR_TEXT)
            if not issubclass(executor_class, NNExecutor):
                message = "%r is not an %s class" % (nn_executor, NN_EXECUTOR_TEXT)
                raise RuntimeError(message)
            executor = executor_class.from_config(nn_config)
            return executor

        nn_name = nn_config.get(NN_NAME_KEY)
        if nn_name is not None:
            nn_name = get_string_field(nn_config, NN_NAME_KEY, NN_NAME_TEXT)
        nn_version = nn_config.get(NN_VERSION_KEY)
        if nn_version is not None:
            nn_version = get_string_field(nn_config, NN_VERSION_KEY, NN_VERSION_TEXT)
        if nn_name is not None:
            hub = NNHub.get_instance()
            executor = hub.get_model_executor(nn_name, nn_version)
            if executor is not None:
                return executor

        message = "can not create executor for NN config"
        if nn_name is not None:
            message += "; model: %r" % nn_name
            if nn_version is not None:
                message += ", version: %r" % nn_version
        raise RuntimeError(message)


class LLMExecutor(NNExecutor, ABC):
    """
    Base Executor for LLM.
    """

    @classmethod
    def from_config(cls, nn_config: Union[str, dict]) -> "LLMExecutor":
        """
        Implement distribution logic for LLM, since we only support Huggingface Decode Only models for now,
        it is directly point to HFDecodeOnlyExecutor. Will use the hub management functions later on.
        """
        from nn4k.executor.huggingface.hf_decode_only_executor import (
            HFDecodeOnlyExecutor,
        )

        return HFDecodeOnlyExecutor.from_config(nn_config)

    def execute_sft(self, args=None, callbacks=None, **kwargs):
        """
        The entry point of SFT execution in a certain pod.
        """
        raise NotImplementedError(f"{self.__class__.__name__} does not support SFT.")

    def execute_rl_tuning(self, args=None, callbacks=None, **kwargs):
        """
        The entry point of SFT execution in a certain pod.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support RL-Tuning."
        )


@dataclass
class NNModelArgs:
    """
    Base NN4K-supported model definition and load related args.
    """

    nn_name: Optional[str] = field(
        default=None,
        metadata={"help": ("NN4K model name")},
    )
    nn_version: Optional[str] = field(
        default="default",
        metadata={"help": ("NN4K model version, by default is 'default'")},
    )
    nn_model_path: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "model path dir, could be delivered by user or get managed in Hub."
            )
        },
    )
    nn_device: Optional[str] = field(
        default="auto", metadata={"help": ("device to use to load model")}
    )

    def __post_init__(self):
        assert (
            self.nn_name is not None or self.nn_model_path is not None
        ), "either nn_name or nn_model_path has to be provided"


@dataclass
class NNAdapterModelArgs(NNModelArgs):
    """
    One should use this args dataclass to enable adapter models.
    """

    adapter_name: str = field(
        default=None,
        metadata={
            "help": "adapter name. Should be provided if you want to sft or load a adapter model."
        },
    )
    adapter_version: str = field(
        default="auto",
        metadata={
            "help": "adapter is designed to get managed by versions, by default is 'latest'"
        },
    )
    adapter_type: str = field(
        default="lora", metadata={"help": "adapter type, lora by default."}
    )
    adapter_path: str = field(
        default=None,
        metadata={
            "help": "adapter weight and config path, could be delivered by user or get managed in Hub."
        },
    )
    adapter_config: Optional[dict] = field(
        default=None,
        metadata={
            "help": "Only necessary if you want to init a new adapter model and train from scratch or resume"
            "from a checkpoint (in this case, should be the same as the previous adapter_config)."
            "Values are the same as peft config init args."
        },
    )

    def __post_init__(self):
        super().__post_init__()


@dataclass
class NNInferenceArgs:
    max_input_length: Optional[int] = field(
        default=None,
        metadata={
            "help": "Controls the maximum length to use by one of the truncation/padding parameters. "
            "In HuggingFace executors, known as max_length in tokenize callable function config."
        },
    )
    max_output_length: Optional[int] = field(
        default=None,
        metadata={
            "help": "The maximum numbers of tokens to generate. In HuggingFace executors, this arg will be tread as "
            "max_new_tokens."
        },
    )
    return_input_text: Optional[bool] = field(
        default=False,
        metadata={"help": "Whether return input texts together with output texts."},
    )
    stop_sequence: Optional[str] = field(
        default=None,
        metadata={
            "help": "Generation will stop when stop sequence encountered in the output."
        },
    )
    do_sample: bool = field(
        default=False,
        metadata={
            "help": "If false, generation will be in greedy search mode, otherwise will sampling the probable tokens."
        },
    )
    temperature: float = field(
        default=1.0,
        metadata={"help": "The creativity and diversity of the text generated."},
    )
    top_k: Optional[int] = field(
        default=50,
        metadata={
            "help": "In nucleus sampling, model will only sampling the tokens with the highest top_p(percentage) "
            "probability"
        },
    )
    top_p: Optional[float] = field(
        default=1.0,
        metadata={
            "help": "In nucleus sampling, model will only sampling the tokens with the highest top_k(count) probability"
        },
    )
    repetition_penalty: Optional[float] = field(
        default=1.0,
        metadata={"help": "By default 1.0 means no penalty."},
    )

    generate_config: dict = field(
        default_factory=lambda: {},
        metadata={"help": "Config dict that will be use in model generation"},
    )

    tokenize_return_tensors: str = field(
        default="pt",
        metadata={
            "help": "Tokenizer return type, will be merged into tokenize_config and pass into tokenize function"
        },
    )
    tokenize_config: dict = field(
        default_factory=lambda: {},
        metadata={
            "help": "Tokenize function config, will be pass into tokenize function"
        },
    )

    decode_config: dict = field(
        default_factory=lambda: {},
        metadata={"help": "Configs to be pass into tokenizer.decode fucntion"},
    )

    def update_if_not_none(self, from_key, to_dict, to_key=None):
        to_key = to_key or from_key
        from_value = self.__getattribute__(from_key)
        value_in_to_dict = self.__getattribute__(to_dict).get(to_key, None)
        if value_in_to_dict is None and from_value is not None:
            self.__getattribute__(to_dict)[to_key] = from_value

    def __post_init__(self):
        # merging generation args
        self.update_if_not_none("max_output_length", "generate_config")
        self.update_if_not_none("do_sample", "generate_config")
        self.update_if_not_none("temperature", "generate_config")
        self.update_if_not_none("top_k", "generate_config")
        self.update_if_not_none("top_p", "generate_config")
        self.update_if_not_none("repetition_penalty", "generate_config")

        # merging tokenize args
        self.update_if_not_none("max_input_length", "tokenize_config", "max_length")
        self.update_if_not_none(
            "tokenize_return_tensors", "tokenize_config", "return_tensors"
        )



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import importlib
from typing import Tuple


def split_module_class_name(name: str, text: str) -> Tuple[str, str]:
    """
    Split `name` as module name and class name pair.

    :param name: fully qualified class name, e.g. ``foo.bar.MyClass``
    :type name: str
    :param text: describe the kind of the class, used in the exception message
    :type text: str
    :rtype: Tuple[str, str]
    :raises RuntimeError: if `name` is not a fully qualified class name
    """
    i = name.rfind(".")
    if i == -1:
        message = "invalid %s class name: %s" % (text, name)
        raise RuntimeError(message)
    module_name = name[:i]
    class_name = name[i + 1 :]
    return module_name, class_name


def dynamic_import_class(name: str, text: str):
    """
    Import the class specified by `name` dyanmically.

    :param name: fully qualified class name, e.g. ``foo.bar.MyClass``
    :type name: str
    :param text: describe the kind of the class, use in the exception message
    :type text: str
    :raises RuntimeError: if `name` is not a fully qualified class name, or
                          the class is not in the module specified by `name`
    :raises ModuleNotFoundError: the module specified by `name` is not found
    """
    module_name, class_name = split_module_class_name(name, text)
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name, None)
    if class_ is None:
        message = "class %r not found in module %r" % (class_name, module_name)
        raise RuntimeError(message)
    if not isinstance(class_, type):
        message = "%r is not a class" % (name,)
        raise RuntimeError(message)
    return class_



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os

import safetensors
import torch
from torch.utils.data import DataLoader, Dataset, RandomSampler, SequentialSampler
from packaging import version

from peft import PeftModel
from transformers import PretrainedConfig, Trainer, __version__
from transformers.integrations import is_deepspeed_available
from transformers.modeling_utils import load_sharded_checkpoint
from transformers.trainer import logger
from transformers.utils import (
    ADAPTER_SAFE_WEIGHTS_NAME,
    ADAPTER_WEIGHTS_NAME,
    CONFIG_NAME,
    SAFE_WEIGHTS_INDEX_NAME,
    SAFE_WEIGHTS_NAME,
    WEIGHTS_INDEX_NAME,
    WEIGHTS_NAME,
    is_accelerate_available,
    is_peft_available,
    is_sagemaker_mp_enabled,
)

if is_accelerate_available():
    from accelerate import Accelerator, skip_first_batches
    from accelerate import __version__ as accelerate_version
    from accelerate.utils import (
        DistributedDataParallelKwargs,
        GradientAccumulationPlugin,
        load_fsdp_model,
        load_fsdp_optimizer,
        save_fsdp_model,
        save_fsdp_optimizer,
    )

    DATA_SAMPLERS = [RandomSampler]
    if version.parse(accelerate_version) > version.parse("0.23.0"):
        from accelerate.data_loader import SeedableRandomSampler

        DATA_SAMPLERS += [SeedableRandomSampler]

    if is_deepspeed_available():
        from accelerate.utils import DeepSpeedSchedulerWrapper


class NNHFTrainer(Trainer):
    """
    only trying to fix resume checkpoint for lora adapter, will be replaced by using Trainer when the bug is
    fixed in huggingface trainer. The PR is offered: https://github.com/huggingface/transformers/pull/28547
    """

    def _load_from_checkpoint(self, resume_from_checkpoint, model=None):
        # the following code only trying to fix resuming checkpoint for adapter model(Peft)
        if model is None:
            model = self.model

        if not (is_peft_available() and isinstance(model, PeftModel)):
            return super()._load_from_checkpoint(resume_from_checkpoint, model)

        adapter_name_path = ""
        if isinstance(model, PeftModel):
            adapter_name_path = (
                model.active_adapter
                if model.active_adapter not in ["default", None]
                else ""
            )

        config_file = os.path.join(resume_from_checkpoint, CONFIG_NAME)
        adapter_weights_file = os.path.join(
            resume_from_checkpoint, adapter_name_path, ADAPTER_WEIGHTS_NAME
        )
        adapter_safe_weights_file = os.path.join(
            resume_from_checkpoint, adapter_name_path, ADAPTER_SAFE_WEIGHTS_NAME
        )
        weights_file = os.path.join(resume_from_checkpoint, WEIGHTS_NAME)
        weights_index_file = os.path.join(resume_from_checkpoint, WEIGHTS_INDEX_NAME)
        safe_weights_file = os.path.join(resume_from_checkpoint, SAFE_WEIGHTS_NAME)
        safe_weights_index_file = os.path.join(
            resume_from_checkpoint, SAFE_WEIGHTS_INDEX_NAME
        )

        if not any(
            os.path.isfile(f)
            for f in [
                weights_file,
                safe_weights_file,
                weights_index_file,
                safe_weights_index_file,
                os.path.join(adapter_weights_file),
                os.path.join(adapter_safe_weights_file),
            ]
        ):
            raise ValueError(
                f"Can't find a valid checkpoint at {resume_from_checkpoint}"
            )

        logger.info(f"Loading model from {resume_from_checkpoint}.")

        if os.path.isfile(config_file):
            config = PretrainedConfig.from_json_file(config_file)
            checkpoint_version = config.transformers_version
            if checkpoint_version is not None and checkpoint_version != __version__:
                logger.warning(
                    f"You are resuming training from a checkpoint trained with {checkpoint_version} of "
                    f"Transformers but your current version is {__version__}. This is not recommended and could "
                    "yield to errors or unwanted behaviors."
                )

        if os.path.isfile(weights_file) or os.path.isfile(safe_weights_file):
            # If the model is on the GPU, it still works!
            if is_sagemaker_mp_enabled():
                if os.path.isfile(
                    os.path.join(resume_from_checkpoint, "user_content.pt")
                ):
                    # If the 'user_content.pt' file exists, load with the new smp api.
                    # Checkpoint must have been saved with the new smp api.
                    import smdistributed.modelparallel.torch as smp

                    smp.resume_from_checkpoint(
                        path=resume_from_checkpoint,
                        tag=WEIGHTS_NAME,
                        partial=False,
                        load_optimizer=False,
                    )
                else:
                    # If the 'user_content.pt' file does NOT exist, load with the old smp api.
                    # Checkpoint must have been saved with the old smp api.
                    if hasattr(self.args, "fp16") and self.args.fp16 is True:
                        logger.warning(
                            "Enabling FP16 and loading from smp < 1.10 checkpoint together is not suppported."
                        )
                    state_dict = torch.load(weights_file, map_location="cpu")
                    # Required for smp to not auto-translate state_dict from hf to smp (is already smp).
                    state_dict["_smp_is_partial"] = False
                    load_result = model.load_state_dict(state_dict, strict=True)
                    # release memory
                    del state_dict
            elif self.is_fsdp_enabled:
                load_fsdp_model(
                    self.accelerator.state.fsdp_plugin,
                    self.accelerator,
                    model,
                    resume_from_checkpoint,
                )
            else:
                # We load the model state dict on the CPU to avoid an OOM error.
                if self.args.save_safetensors and os.path.isfile(safe_weights_file):
                    state_dict = safetensors.torch.load_file(
                        safe_weights_file, device="cpu"
                    )
                else:
                    state_dict = torch.load(weights_file, map_location="cpu")

                # workaround for FSDP bug https://github.com/pytorch/pytorch/issues/82963
                # which takes *args instead of **kwargs
                load_result = model.load_state_dict(state_dict, False)
                # release memory
                del state_dict
                self._issue_warnings_after_load(load_result)

        # Load adapters following PR # 24096
        elif is_peft_available() and isinstance(model, PeftModel):
            # If train a model using PEFT & LoRA, assume that adapter have been saved properly.
            if hasattr(model, "active_adapter") and hasattr(model, "load_adapter"):
                adapter_model_path = os.path.join(
                    resume_from_checkpoint, adapter_name_path
                )
                if os.path.exists(adapter_model_path):
                    model.load_adapter(
                        adapter_model_path, model.active_adapter, is_trainable=True
                    )
                else:
                    logger.warning(
                        "The intermediate checkpoints of PEFT may not be saved correctly, "
                        f"consider using a custom callback to save {ADAPTER_WEIGHTS_NAME} in corresponding saving folders. "
                        "Check some examples here: https://github.com/huggingface/peft/issues/96"
                    )
            else:
                logger.warning(
                    "Could not load adapter model, make sure to have `peft>=0.3.0` installed"
                )
        else:
            # We load the sharded checkpoint
            load_result = load_sharded_checkpoint(
                model,
                resume_from_checkpoint,
                strict=is_sagemaker_mp_enabled(),
                prefer_safe=self.args.save_safetensors,
            )
            if not is_sagemaker_mp_enabled():
                self._issue_warnings_after_load(load_result)



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from nn4k.invoker.base import NNInvoker


def main():
    # example for local sft
    # NNInvoker.from_config("local_sft.json5").local_sft()

    # example for local inference
    invoker = NNInvoker.from_config("local_infer.json5")
    answer = invoker.local_inference(
        "What could LLM do for human?",
        tokenize_config={"padding": True},
        delete_heading_new_lines=True,
    )
    # doing so to avoid load model everytime. You could hold a invoker, which has alreday load the model at the first time.
    answer2 = invoker.local_inference(
        "What could LLM do for a programmer",
        tokenize_config={"padding": True},
        delete_heading_new_lines=True,
    )


if __name__ == "__main__":
    main()



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from typing import Optional

from nn4k.invoker import NNInvoker


class OpenAIInvoker(NNInvoker):
    def __init__(self, nn_config: dict):
        super().__init__(nn_config)

        import openai
        from nn4k.consts import NN_OPENAI_MODEL_NAME_KEY, NN_OPENAI_MODEL_NAME_TEXT
        from nn4k.consts import NN_OPENAI_API_KEY_KEY, NN_OPENAI_API_KEY_TEXT
        from nn4k.consts import NN_OPENAI_API_BASE_KEY, NN_OPENAI_API_BASE_TEXT
        from nn4k.consts import NN_OPENAI_MAX_TOKENS_KEY, NN_OPENAI_MAX_TOKENS_TEXT
        from nn4k.consts import NN_OPENAI_ORGANIZATION_KEY, NN_OPENAI_ORGANIZATION_TEXT
        from nn4k.utils.config_parsing import get_string_field
        from nn4k.utils.config_parsing import get_positive_int_field

        self.openai_model_name = get_string_field(
            self.init_args, NN_OPENAI_MODEL_NAME_KEY, NN_OPENAI_MODEL_NAME_TEXT
        )
        self.openai_api_key = get_string_field(
            self.init_args, NN_OPENAI_API_KEY_KEY, NN_OPENAI_API_KEY_TEXT
        )
        self.openai_api_base = get_string_field(
            self.init_args, NN_OPENAI_API_BASE_KEY, NN_OPENAI_API_BASE_TEXT
        )
        self.openai_max_tokens = self.init_args.get(NN_OPENAI_MAX_TOKENS_KEY)
        if self.openai_max_tokens is not None:
            self.openai_max_tokens = get_positive_int_field(
                self.init_args, NN_OPENAI_MAX_TOKENS_KEY, NN_OPENAI_MAX_TOKENS_TEXT
            )
        self.openai_organization = self.init_args.get(NN_OPENAI_ORGANIZATION_KEY)
        if self.openai_organization is not None:
            self.openai_organization = get_string_field(
                self.init_args, NN_OPENAI_ORGANIZATION_KEY, NN_OPENAI_ORGANIZATION_TEXT
            )

        if self._is_legacy_openai_api:
            self.client = None
            openai.api_key = self.openai_api_key
            openai.api_base = self.openai_api_base
            openai.organization = self.openai_organization
        else:
            self.client = openai.OpenAI(
                api_key=self.openai_api_key,
                base_url=self.openai_api_base,
                organization=self.openai_organization,
            )

    @classmethod
    def from_config(cls, nn_config: dict) -> "OpenAIInvoker":
        invoker = cls(nn_config)
        return invoker

    @property
    def _is_legacy_openai_api(self):
        import openai

        return openai.__version__.startswith("0.")

    def _create_prompt(self, input, **kwargs):
        if isinstance(input, list):
            prompt = input
        else:
            prompt = [{"role": "user", "content": input}]
        return prompt

    def _create_completion(self, input, prompt, max_output_length, **kwargs):
        import openai

        if self._is_legacy_openai_api:
            completion = openai.ChatCompletion.create(
                model=self.openai_model_name,
                messages=prompt,
                max_tokens=max_output_length,
            )
        else:
            completion = self.client.chat.completions.create(
                model=self.openai_model_name,
                messages=prompt,
                max_tokens=max_output_length,
            )
        return completion

    def _create_output(self, input, prompt, completion, **kwargs):
        output = [choice.message.content for choice in completion.choices]
        return output

    def _get_completion(self, input, max_output_length, **kwargs):
        prompt = self._create_prompt(input, **kwargs)
        completion = self._create_completion(input, prompt, max_output_length, **kwargs)
        output = self._create_output(input, prompt, completion, **kwargs)
        return output

    def _get_embeddings(self, input, **kwargs):
        import openai

        if isinstance(input, list):
            inputs = input
        else:
            inputs = [input]

        if self._is_legacy_openai_api:
            response = openai.Embedding.create(
                model=self.openai_model_name,
                input=inputs,
            )
        else:
            response = self.client.embeddings.create(
                model=self.openai_model_name,
                input=inputs,
            )

        embeddings = [emb.embedding for emb in response.data]
        return embeddings

    def remote_inference(
        self,
        input,
        type: Optional[str] = None,
        max_output_length: Optional[int] = None,
        **kwargs
    ):
        if type == "Embedding":
            output = self._get_embeddings(input, **kwargs)
        else:
            if max_output_length is None:
                max_output_length = self.openai_max_tokens
            output = self._get_completion(input, max_output_length, **kwargs)
        return output



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

NN_NAME_KEY = "nn_name"
NN_NAME_TEXT = "NN model name"

NN_VERSION_KEY = "nn_version"
NN_VERSION_TEXT = "NN model version"
NN_VERSION_DEFAULT = "default"

NN_INVOKER_KEY = "nn_invoker"
NN_INVOKER_TEXT = "NN invoker"

NN_EXECUTOR_KEY = "nn_executor"
NN_EXECUTOR_TEXT = "NN executor"

NN_DEVICE_KEY = "nn_device"

NN_OPENAI_MODEL_NAME_KEY = NN_NAME_KEY
NN_OPENAI_MODEL_NAME_TEXT = "openai model name"

NN_OPENAI_API_KEY_KEY = "openai_api_key"
NN_OPENAI_API_KEY_TEXT = "openai api key"

NN_OPENAI_API_BASE_KEY = "openai_api_base"
NN_OPENAI_API_BASE_TEXT = "openai api base"

NN_OPENAI_MAX_TOKENS_KEY = "openai_max_tokens"
NN_OPENAI_MAX_TOKENS_TEXT = "openai max tokens"

NN_OPENAI_ORGANIZATION_KEY = "openai_organization"
NN_OPENAI_ORGANIZATION_TEXT = "openai organization"

NN_OPENAI_GPT4_PREFIX = "gpt-4"
NN_OPENAI_GPT35_PREFIX = "gpt-3.5"
NN_OPENAI_EMBEDDING_PREFIX = "text-embedding"

NN_LOCAL_HF_MODEL_CONFIG_FILE = "config.json"
NN_LOCAL_SENTENCE_TRANSFORMERS_CONFIG_FILE = "config_sentence_transformers.json"

NN_HUB_INFER_URL_KEY = "hub_infer_url"
NN_HUB_INFER_URL_TEXT = "hub inference url"



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from pathlib import Path
from typing import Any, Union


def preprocess_config(nn_config: Union[str, dict]) -> dict:
    """
    Preprocess config `nn_config` into a dictionary.

    * If `nn_config` is already a dictionary, return it as is.

    * If `nn_config` is a string, decode it as a JSON or JSON5 file.

    :param nn_config: config to be preprocessed
    :type nn_config: str or dict
    :return: `nn_config` or `nn_config` decoded as JSON or JSON5
    :rtype: dict
    :raises ValueError: if cannot decode config file specified by
                        `nn_config` as JSON or JSON5
    """
    try:
        if isinstance(nn_config, dict):
            return nn_config
        elif isinstance(nn_config, str):
            from nn4k.utils.args_utils import ArgsUtils

            return ArgsUtils.load_config_dict_from_file(nn_config)
        else:
            raise ValueError(
                f"nn_config could be dict or str, {type(nn_config)} is not yet supported."
            )
    except:
        raise ValueError("cannot decode config file")


def get_field(nn_config: dict, name: str, text: str) -> Any:
    """
    Get the value of the field specified by `name` from the configuration
    dictionary `nn_config`.

    :param str name: name of the field
    :param str name: descriptive text of the name of the field
    :return: value of the field
    :rtype: Any
    :raises ValueError: if the field is not specified in `nn_config`
    """
    value = nn_config.get(name)
    if value is None:
        message = "%s %r not found" % (text, name)
        raise ValueError(message)
    return value


def get_string_field(nn_config: dict, name: str, text: str) -> str:
    """
    Get the value of the string field specified by `name` from the
    configuration dictionary `nn_config`.

    :param str name: name of the field
    :param str name: descriptive text of the name of the field
    :return: value of the field
    :rtype: str
    :raises ValueError: if the field is not specified in `nn_config`
    :raises TypeError: if the value of the field is not a string
    """
    value = get_field(nn_config, name, text)
    if not isinstance(value, str):
        message = "%s %r must be string; " % (text, name)
        message += "%r is invalid" % (value,)
        raise TypeError(message)
    return value


def get_int_field(nn_config: dict, name: str, text: str) -> int:
    """
    Get the value of the integer field specified by `name` from the
    configuration dictionary `nn_config`.

    :param str name: name of the field
    :param str name: descriptive text of the name of the field
    :return: value of the field
    :rtype: int
    :raises ValueError: if the field is not specified in `nn_config`
    :raises TypeError: if the value of the field is not an integer
    """
    value = get_field(nn_config, name, text)
    if not isinstance(value, int):
        message = "%s %r must be integer; " % (text, name)
        message += "%r is invalid" % (value,)
        raise TypeError(message)
    return value


def get_positive_int_field(nn_config: dict, name: str, text: str) -> int:
    """
    Get the value of the positive integer field specified by `name`
    from the configuration dictionary `nn_config`.

    :param str name: name of the field
    :param str name: descriptive text of the name of the field
    :return: value of the field
    :rtype: int
    :raises ValueError: if the field is not specified in `nn_config`, or the
                        value of the field is not a positive integer
    :raises TypeError: if the value of the field is not an integer
    """
    value = get_int_field(nn_config, name, text)
    if value <= 0:
        message = "%s %r must be positive integer; " % (text, name)
        message += "%r is invalid" % (value,)
        raise ValueError(message)
    return value



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import copy
from abc import ABC, abstractmethod
from enum import Enum
from typing import Union

from nn4k.utils.class_importing import dynamic_import_class

from nn4k.executor import LLMExecutor
from nn4k.utils.args_utils import ArgsUtils


class SubmitMode(Enum):
    K8s = "k8s"
    Docker = "docker"
    Local = "local"


class NNInvoker(ABC):
    """
    Invoking Entry Interfaces for NN Models.
    One NNInvoker object is for one NN Model.
    - Interfaces starting with "submit_" means submitting a batch task to a remote execution engine.
    - Interfaces starting with "remote_" means querying a remote service for some results.
    - Interfaces starting with "local_"  means running something locally.
            Must call `warmup_local_model` before calling any local_xxx interface.
    """

    def __init__(self, init_args: dict, **kwargs):
        self._init_args = init_args
        self._kwargs = kwargs
        self.inference_warmed_up = False

    @property
    def init_args(self):
        """
        Return the `init_args` passed to the invoker constructor.
        """
        return self._init_args

    @property
    def kwargs(self):
        """
        Return the `kwargs` passed to the invoker constructor.
        """
        return self._kwargs

    @classmethod
    @abstractmethod
    def from_config(cls, nn_config: Union[str, dict]) -> "NNInvoker":
        """
        Create an NN invoker instance from `nn_config`.

        This method is abstract, derived class must override it by either
        creating invoker instances or implementating dispatch logic.

        :param nn_config: config to use, can be dictionary or path to a JSON file
        :type nn_config: str or dict
        :rtype: NNInvoker
        :raises RuntimeError: if the NN config is not recognized
        """
        from nn4k.nnhub import NNHub
        from nn4k.consts import NN_NAME_KEY, NN_NAME_TEXT
        from nn4k.consts import NN_VERSION_KEY, NN_VERSION_TEXT
        from nn4k.consts import NN_INVOKER_KEY, NN_INVOKER_TEXT
        from nn4k.utils.config_parsing import preprocess_config
        from nn4k.utils.config_parsing import get_string_field

        nn_config = preprocess_config(nn_config)
        nn_invoker = nn_config.get(NN_INVOKER_KEY)
        if nn_invoker is not None:
            nn_invoker = get_string_field(nn_config, NN_INVOKER_KEY, NN_INVOKER_TEXT)
            invoker_class = dynamic_import_class(nn_invoker, NN_INVOKER_TEXT)
            if not issubclass(invoker_class, NNInvoker):
                message = "%r is not an %s class" % (nn_invoker, NN_INVOKER_TEXT)
                raise RuntimeError(message)
            invoker = invoker_class.from_config(nn_config)
            return invoker

        hub = NNHub.get_instance()
        invoker = hub.get_invoker(nn_config)
        if invoker is not None:
            return invoker

        nn_name = nn_config.get(NN_NAME_KEY)
        if nn_name is not None:
            nn_name = get_string_field(nn_config, NN_NAME_KEY, NN_NAME_TEXT)
        nn_version = nn_config.get(NN_VERSION_KEY)
        if nn_version is not None:
            nn_version = get_string_field(nn_config, NN_VERSION_KEY, NN_VERSION_TEXT)
        message = "can not create invoker for NN config"
        if nn_name is not None:
            message += "; model: %r" % nn_name
            if nn_version is not None:
                message += ", version: %r" % nn_version
        raise RuntimeError(message)

    def submit_inference(self, submit_mode: SubmitMode = SubmitMode.K8s):
        """
        Submit remote batch inference execution.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support batch inference."
        )

    def remote_inference(self, input, **kwargs):
        """
        Inference via existing remote services.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support remote inference."
        )

    def local_inference(self, data, **kwargs):
        """
        Implement local inference in derived invoker classes.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support local inference."
        )

    def warmup_local_model(self):
        """
        Implement local model warming up logic in derived invoker classes.
        """
        pass


class LLMInvoker(NNInvoker):
    def submit_sft(self, submit_mode: SubmitMode = SubmitMode.K8s):
        """
        Submit remote SFT execution.
        """
        raise NotImplementedError(f"{self.__class__.__name__} does not support SFT.")

    def local_sft(self, args: dict = None):
        sft_args = ArgsUtils.update_args(self.init_args, args)

        from nn4k.executor import LLMExecutor

        LLMExecutor.from_config(sft_args).execute_sft()

    def submit_rl_tuning(self, submit_mode: SubmitMode = SubmitMode.K8s):
        """
        Submit remote RL-Tuning execution.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support RL-Tuning."
        )

    def local_inference(self, data, **kwargs):
        """
        Implement local inference for local invoker.
        """
        args = ArgsUtils.handle_dict_config(kwargs)

        if not self.inference_warmed_up:
            print(
                "warming up the model for inference, only happen for the first time..."
            )
            self.warmup_local_model()
            self.inference_warmed_up = True
            print("inference model is warmed up")

        return self._nn_executor.inference(inputs=data, **args)

    def warmup_local_model(self):
        """
        Implement local model warming up logic for local invoker.
        """
        nn_config = self.init_args

        from nn4k.nnhub import NNHub
        from nn4k.consts import NN_NAME_KEY, NN_NAME_TEXT
        from nn4k.consts import NN_VERSION_KEY, NN_VERSION_TEXT
        from nn4k.utils.config_parsing import get_string_field
        from transformers import HfArgumentParser
        from nn4k.executor import NNModelArgs

        parser = HfArgumentParser(NNModelArgs)
        model_args: NNModelArgs
        model_args, *_ = parser.parse_dict(self.init_args, allow_extra_keys=True)

        from nn4k.consts import NN_EXECUTOR_KEY, NN_EXECUTOR_TEXT

        nn_executor = nn_config.get(NN_EXECUTOR_KEY)
        if nn_executor is not None:
            from nn4k.executor import NNExecutor

            executor_class = dynamic_import_class(nn_executor, NN_EXECUTOR_TEXT)
            if not issubclass(executor_class, NNExecutor):
                message = "%r is not an %s class" % (nn_executor, NN_EXECUTOR_TEXT)
                raise RuntimeError(message)
            executor = executor_class.from_config(nn_config)
        else:
            nn_name = get_string_field(self.init_args, NN_NAME_KEY, NN_NAME_TEXT)
            nn_version = self.init_args.get(NN_VERSION_KEY)
            if nn_version is not None:
                nn_version = get_string_field(
                    self.init_args, NN_VERSION_KEY, NN_VERSION_TEXT
                )
            hub = NNHub.get_instance()
            executor = hub.get_model_executor(nn_name, nn_version)

        if executor is None:
            message = "model %r version %r " % (
                model_args.nn_name,
                model_args.nn_version,
            )
            message += "is not found in the model hub, you should provide a valid nn_executor class path"
            raise RuntimeError(message)
        self._nn_executor: LLMExecutor = executor
        self._nn_executor.load_model(mode="inference")
        self._nn_executor.warmup_inference()
        self.inference_warmed_up = True

    @classmethod
    def from_config(cls, nn_config: dict) -> "LLMInvoker":
        """
        Create an LLMInvoker instance from `nn_config`.
        """
        invoker = cls(nn_config)
        return invoker



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os


def is_openai_invoker(nn_config: dict) -> bool:
    """
    Check whether `nn_config` specifies OpenAI invoker.

    :type nn_config: dict
    :rtype: bool
    """
    from nn4k.consts import NN_NAME_KEY, NN_NAME_TEXT
    from nn4k.consts import NN_OPENAI_API_KEY_KEY
    from nn4k.consts import NN_OPENAI_API_BASE_KEY
    from nn4k.consts import NN_OPENAI_MAX_TOKENS_KEY
    from nn4k.consts import NN_OPENAI_ORGANIZATION_KEY
    from nn4k.consts import NN_OPENAI_GPT4_PREFIX
    from nn4k.consts import NN_OPENAI_GPT35_PREFIX
    from nn4k.consts import NN_OPENAI_EMBEDDING_PREFIX
    from nn4k.utils.config_parsing import get_string_field

    nn_name = nn_config.get(NN_NAME_KEY)
    if nn_name is not None:
        nn_name = get_string_field(nn_config, NN_NAME_KEY, NN_NAME_TEXT)
        if (
            nn_name.startswith(NN_OPENAI_GPT4_PREFIX)
            or nn_name.startswith(NN_OPENAI_GPT35_PREFIX)
            or nn_name.startswith(NN_OPENAI_EMBEDDING_PREFIX)
        ):
            return True
    keys = (
        NN_OPENAI_API_KEY_KEY,
        NN_OPENAI_API_BASE_KEY,
        NN_OPENAI_MAX_TOKENS_KEY,
        NN_OPENAI_ORGANIZATION_KEY,
    )
    for key in keys:
        if key in nn_config:
            return True
    return False


def is_local_invoker(nn_config: dict) -> bool:
    """
    Check whether `nn_config` specifies local invoker.

    :type nn_config: dict
    :rtype: bool
    """
    from nn4k.consts import NN_NAME_KEY, NN_NAME_TEXT
    from nn4k.consts import NN_LOCAL_HF_MODEL_CONFIG_FILE
    from nn4k.utils.config_parsing import get_string_field

    nn_name = nn_config.get(NN_NAME_KEY)
    if nn_name is not None:
        nn_name = get_string_field(nn_config, NN_NAME_KEY, NN_NAME_TEXT)
        if os.path.isdir(nn_name):
            file_path = os.path.join(nn_name, NN_LOCAL_HF_MODEL_CONFIG_FILE)
            if os.path.isfile(file_path):
                return True
    return False


def is_hub_invoker(nn_config: dict) -> bool:
    """
    Check whether `nn_config` specifies hub invoker.

    :type nn_config: dict
    :rtype: bool
    """
    from nn4k.consts import NN_HUB_INFER_URL_KEY, NN_HUB_INFER_URL_TEXT
    from nn4k.utils.config_parsing import get_string_field

    hub_infer_url = nn_config.get(NN_HUB_INFER_URL_KEY)
    if hub_infer_url is not None:
        hub_infer_url = get_string_field(
            nn_config, NN_HUB_INFER_URL_KEY, NN_HUB_INFER_URL_TEXT
        )
        return True

    return False



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
from typing import List

from nn4k.utils.io.file_utils import FileUtils

EXTENSION_TYPE = {"csv": "csv", "json": "json", "jsonl": "json", "txt": "text"}


class DatasetUtils:
    @staticmethod
    def auto_dataset(input_path, split="train", transform_fn=None, dataset_map_fn=None):
        """
        Args:
            input_path: dataset pash, support local file path or dir, if dir is used, make sure all files within the dir
                has the same file extension
            split: data split of dataset, see dataset doc for more info.
            transform_fn: transform function for dataset
            dataset_map_fn: dataset map function
        """
        dataset_dir = input_path
        file_extension = None
        data_files: List[str] = []
        if os.path.isdir(input_path):  # support directory
            for file_name in os.listdir(input_path):
                data_files.append(os.path.join(input_path, file_name))
                if file_extension is None:
                    file_extension = EXTENSION_TYPE.get(
                        FileUtils.get_extension(file_name), None
                    )
                else:
                    assert file_extension == EXTENSION_TYPE.get(
                        FileUtils.get_extension(file_name), None
                    ), "file type does not match."
        elif os.path.isfile(dataset_dir):  # support single file
            data_files.append(dataset_dir)
            file_extension = EXTENSION_TYPE.get(
                FileUtils.get_extension(dataset_dir), None
            )
        else:
            raise ValueError("File not found.")

        from datasets import load_dataset

        dataset = load_dataset(
            file_extension,
            data_files=data_files,
            split=split,
        )
        if transform_fn is not None:
            dataset.set_transform(transform_fn)

        if dataset_map_fn is not None:
            dataset = dataset.map(dataset_map_fn)

        return dataset



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from dataclasses import dataclass, field
from typing import Optional

from transformers import TrainingArguments

from nn4k.executor import NNAdapterModelArgs
from nn4k.executor.base import NNInferenceArgs


@dataclass
class HFModelArgs(NNAdapterModelArgs):
    """
    Huggingface Model is designed to support adapter models such as lora, therefore should inherit from
    NNAdapterModelArgs dataclass
    """

    torch_dtype: Optional[str] = field(
        default="auto",
        metadata={
            "help": (
                "Override the default `torch.dtype` and load the model under this dtype. If `auto` is passed, the "
                "dtype will be automatically derived from the model's weights."
            )
        },
    )
    qlora_bits_and_bytes_config: Optional[dict] = field(
        default=None,
        metadata={
            "help": "Quantization configs to load qlora, "
            "same as :class:`transformers.utils.quantization_config.BitsAndBytesConfig`"
        },
    )
    trust_remote_code: bool = field(
        default=True,
        metadata={
            "help": "Whether or not to allow for custom models defined on the Hub in their own modeling files."
        },
    )
    from_tf: bool = field(
        default=False,
        metadata={
            "help": " Load the model weights from a TensorFlow checkpoint save file, default to False"
        },
    )
    padding_side: Optional[str] = field(
        default=None,
        metadata={
            "help": "Padding side of the tokenizer when padding batch inputs",
            "choices": [None, "left", "right"],
        },
    )

    def __post_init__(self):
        super().__post_init__()
        # for hf models, if model path has higher priority then name, since you don't need to download the model(or
        # from cache) again.
        self.pretrained_model_name_or_path = self.nn_model_path or self.nn_name


@dataclass
class HFSftArgs(HFModelArgs, TrainingArguments):
    """
    args to use for huggingface model sft task
    """

    train_dataset_path: Optional[str] = field(
        default=None,
        metadata={
            "help": "Should not be None. A file or dir path to train dataset, If a dir path, "
            "all files inside should have the same file extension."
        },
    )
    eval_dataset_path: Optional[str] = field(
        default=None,
        metadata={
            "help": "A file or dir path to eval dataset. If a dir path, all files inside should have the same "
            "file extension. If set, do_eval flag will be set to True"
        },
    )
    max_input_length: int = field(
        default=1024,
        metadata={"help": "max length of input"},
    )
    resume_from_checkpoint: Optional[str] = field(
        default=None,
        metadata={
            "help": "The path to a folder with a valid checkpoint for your model."
        },
    )

    def __post_init__(self):
        HFModelArgs.__post_init__(self)
        TrainingArguments.__post_init__(self)
        assert self.train_dataset_path is not None, "train_dataset_path must be set."
        if self.train_dataset_path and not self.do_train:
            self.do_train = True
            print(
                f"a train_dataset_path is set but do_train flag is not set, automatically set do_train to True"
            )
        if self.eval_dataset_path and not self.do_eval:
            self.do_eval = True
            print(
                f"a eval_dataset_path is set but do_eval flag is not set, automatically set do_eval to True"
            )


@dataclass
class HFInferArgs(NNInferenceArgs):
    delete_heading_new_lines: bool = field(
        default=False,
        metadata={
            "help": "an additional question mark or new line marks sometimes occurs at the beginning of output."
            "Try to get rid of these marks by setting this parameter to True. Different model may have different "
            "behavior, please check the result carefully."
        },
    )

    tokenize_config: dict = field(
        default_factory=lambda: {
            "add_special_tokens": False,
            "padding": False,
            "truncation": False,
        },
        metadata={
            "help": "padding: https://huggingface.co/docs/transformers/pad_truncation#padding-and-truncation"
        },
    )

    decode_config: dict = field(
        default_factory=lambda: {
            "skip_special_tokens": True,
            "clean_up_tokenization_spaces": True,
        },
        metadata={
            "help": "check https://huggingface.co/docs/transformers/main_classes/tokenizer#transformers.PreTrainedTokenizer.__call__"
        },
    )

    def __post_init__(self):
        super().__post_init__()

        # HF specific map
        self.update_if_not_none(
            "max_output_length", "generate_config", "max_new_tokens"
        )
        self.update_if_not_none("max_input_length", "tokenize_config", "max_length")



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import torch
import transformers
from transformers import AutoModelForCausalLM

from nn4k.executor.huggingface import HFModelArgs
from nn4k.executor.huggingface import HFLLMExecutor


class HFDecodeOnlyExecutor(HFLLMExecutor):
    """
    Huggingface decode only default executor, will use AutoModelForCausalLM to load model and
    DataCollatorForSeq2Seq as a default data collator.
    """

    def _hf_model_loader(
        self,
        args: HFModelArgs,
        mode,
        resume_from_checkpoint=False,
        device=None,
        **kwargs,
    ):
        if device is None or "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"

        # load base model
        model_config = self._hf_model_config_loader(args, **kwargs)

        quant_config = None
        if args.adapter_name and args.qlora_bits_and_bytes_config:
            from transformers import BitsAndBytesConfig

            quant_config = BitsAndBytesConfig(**args.qlora_bits_and_bytes_config)

        model_load_args = dict(
            pretrained_model_name_or_path=args.pretrained_model_name_or_path,
            config=model_config,
            quantization_config=quant_config,
            revision=args.nn_version,
            torch_dtype=args.torch_dtype,
            from_tf=args.from_tf,
            trust_remote_code=args.trust_remote_code,
        )

        model = AutoModelForCausalLM.from_pretrained(**model_load_args)

        if quant_config:
            from peft import prepare_model_for_kbit_training

            model = prepare_model_for_kbit_training(model)

        # load adapter model
        if args.adapter_name:
            # provide an adapter_path, means one can load an exist lora adapter and start a new train based on that.
            if args.adapter_path and not resume_from_checkpoint:
                from peft import PeftModel

                # TODO NN4K: Notice: NN4K plan to provide a hub-managed adapter implementation in the near future.
                model = PeftModel.from_pretrained(
                    model=model,
                    model_id=args.adapter_path,
                    adapter_name=args.adapter_name,
                    adapter_version=args.adapter_version,
                    is_trainable=(mode == "train"),
                )
            elif (
                args.adapter_config
            ):  # no adapter_path but adapter_config means train an adapter from scratch
                from peft import get_peft_model
                from peft import LoraConfig

                if args.adapter_type in ["lora", "qlora"]:
                    peft_config = LoraConfig(**args.adapter_config)
                else:
                    raise NotImplementedError(
                        f"adapter_type {args.adapter_type} is not supported in "
                        f"hf_decode_only_executor use lora or qlora instead"
                    )
                model = get_peft_model(
                    model=model,
                    peft_config=peft_config,
                    adapter_name=args.adapter_name,
                    # TODO NN4K: NN4K plan to provide a hub-managed adapter implementation in the
                    #  near future. adapter_version=args.adapter_version,
                )
            else:
                raise ValueError(
                    "You should either provide a adapter_path to load an existing adapter without resume"
                    "a training, or provide a adapter_config to train a adapter from scratch or resume a "
                    "adapter training from checkpoint."
                )
            model.print_trainable_parameters()

        if mode == "inference":
            model.eval()
        model.to(device)

        return model

    def _data_collator(self, return_tensors="pt", **kwargs):
        return transformers.DataCollatorForSeq2Seq(
            self.tokenizer,
            pad_to_multiple_of=8,
            return_tensors=return_tensors,
            padding=True,
        )



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os

from setuptools import setup, find_packages

package_name = "openspg-nn4k"

# version
cwd = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(cwd, "NN4K_VERSION"), "r") as rf:
    version = rf.readline().strip("\n").strip()

# license
license = ""
with open(os.path.join(cwd, "LICENSE"), "r") as rf:
    line = rf.readline()
    while line:
        line = line.strip()
        if line:
            license += "# " + line + "\n"
        else:
            license += "#\n"
        line = rf.readline()

# Generate nn4k.__init__.py
with open(os.path.join(cwd, "nn4k/__init__.py"), "w") as wf:
    content = f"""{license}

__package_name__ = "{package_name}"
__version__ = "{version}"
"""
    wf.write(content)

setup(
    name=package_name,
    version=version,
    description="nn4k",
    url="https://github.com/OpenSPG/openspg",
    packages=find_packages(
        where=".",
        exclude=[
            ".*test.py",
            "*_test.py",
            "*_debug.py",
            "*.txt",
            "tests",
            "tests.*",
            "configs",
            "configs.*",
            "test",
            "test.*",
            "*.tests",
            "*.tests.*",
            "*.pyc",
        ],
    ),
    python_requires=">=3.8",
    install_requires=[
        r.strip()
        for r in open("requirements.txt", "r")
        if not r.strip().startswith("#")
    ],
    include_package_data=True,
    package_data={
        "bin": ["*"],
    },
)



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import requests
import json
from nn4k.invoker.base import NNInvoker
from dataclasses import dataclass, field
from typing import Optional
from transformers import HfArgumentParser


class HubInvoker(NNInvoker):
    """
    Invoking Entry Interfaces for remote inference service.
    """

    def __init__(self, nn_config: dict, **kwargs):
        super().__init__(nn_config, **kwargs)
        parser = HfArgumentParser(HubInvokerArgs)
        self.invoker_args: HubInvokerArgs
        self.invoker_args, *_ = parser.parse_dict(self.init_args, allow_extra_keys=True)

    def _check_input(self, input, **kwargs):
        """
        Filter the input that can not currently be processed.
        """
        if isinstance(input, list):
            raise NotImplementedError(
                f"Hub invoker does not support batch inference at the moment."
            )
        else:
            prompt = input

        return prompt

    def _call_service(self, prompt):
        """
        Call the remote services to do the inference.
        """
        url = self.invoker_args.hub_infer_url
        headers = {"Content-Type": "application/json"}

        data = {"text_input": prompt, "parameters": self.invoker_args.generate_config}

        response = requests.post(url=url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            response_result = response.json()["text_output"]

            if isinstance(response_result, list):
                return response_result
            else:
                newline_pos = response_result.find("\n")
                if newline_pos != -1:
                    return [response_result[newline_pos + 1 :]]
                else:
                    return [response_result]
        else:
            raise NN4KException(
                f"response error No: {response.status_code},  response error text: {response.text}"
            )

    def remote_inference(self, input, **kwargs):
        """
        Inference via existing remote services.
        """
        prompt = self._check_input(input, **kwargs)
        output = self._call_service(prompt)

        return output

    @classmethod
    def from_config(cls, nn_config: dict) -> "HubInvoker":
        """
        Create an HubInvoker instance  from 'nn_config'.
        """
        invoker = cls(nn_config)

        return invoker


@dataclass
class HubInvokerArgs:
    """
    Base HubInvoker-supported related args.
    """

    hub_infer_url: Optional[str] = field(
        default=None, metadata={"help": ("The remote inference service url.")}
    )
    generate_config: Optional[dict] = field(
        default=None, metadata={"help": ("Inference related configs.")}
    )


class NN4KException(Exception):
    """
    Wrapping a NN4K Exception class.
    """

    pass



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
import typing
from abc import abstractmethod
from typing import Optional, Union

from torch.utils.data import Dataset
from transformers import AutoConfig, AutoTokenizer, Trainer

from nn4k.executor import LLMExecutor
from .hf_args import HFInferArgs, HFSftArgs, HFModelArgs
from nn4k.executor.huggingface.nn_hf_trainer import NNHFTrainer
from nn4k.utils.args_utils import ArgsUtils


class HFLLMExecutor(LLMExecutor):
    """
    Base Executor for huggingface models.
    """

    def __init__(self, init_args: dict, **kwargs):
        super().__init__(init_args=init_args, **kwargs)
        # model_model could be either 'train' or 'inference' or model load
        self.model_mode = None

    @classmethod
    def from_config(cls, nn_config: Union[dict]) -> "HFLLMExecutor":
        """
        Create an HFLLMExecutor instance from `nn_config`.
        """
        executor = cls(nn_config)
        return executor

    def execute_sft(self, args: dict = None, callbacks=None, **kwargs):
        args = args or self.init_args

        self.load_model(args=args, mode="train")

        # parse args into HFSftArgs dataclass for more convenient features
        from transformers import HfArgumentParser

        parser = HfArgumentParser(HFSftArgs)
        hf_sft_args: HFSftArgs
        hf_sft_args, *_ = parser.parse_dict(args, allow_extra_keys=True)

        # load checkpoint path if necessary.
        resume_from_checkpoint_path = self._get_last_checkpoint(hf_sft_args)

        # load and map dataset
        train_dataset, eval_dataset = self._init_dataset(hf_sft_args)

        # init trainer
        trainer: Trainer = self._init_trainer(
            train_dataset, eval_dataset, hf_sft_args, callbacks
        )

        # start training
        train_result = trainer.train(resume_from_checkpoint=resume_from_checkpoint_path)

        # save trained model after train complete
        trainer.save_model(hf_sft_args.output_dir)

        # save train metrics
        train_metrics = train_result.metrics
        train_metrics["train_samples_len"] = len(train_dataset)
        trainer.log_metrics("train", train_metrics)
        trainer.save_metrics("train", train_metrics)
        trainer.save_state()

        return self

    def _get_last_checkpoint(self, sft_args: HFSftArgs) -> Optional[str]:  # noqa
        """
        try to find checkpoint in sft_args.output_dir.
        If sft_args.resume_from_checkpoint in ['True', 'true', True, ''], try to return the checkpoint dir with the
        largest checkpoint index. The largest checkpoint dir path will be returned.
        If sft_args.resume_from_checkpoint in [None, 'False', 'false', False], means not necessary to resume from
        checkpoint, None will be returned.
        If sft_args.resume_from_checkpoint is the checkpoint subfolder dir name, the 'output_dir/resume_from_checkpoint'
        path will be returned if exists. Be aware, if the dir does not exist, ValueError will be raised.
        """
        output_dir_contains_file = (
            os.path.isdir(sft_args.output_dir)
            and len(os.listdir(sft_args.output_dir)) > 0
        )

        if sft_args.resume_from_checkpoint in ["True", "true", True, ""]:
            resume_from_checkpoint_bool = True
            if output_dir_contains_file:
                from transformers.trainer_utils import get_last_checkpoint

                resume_from_checkpoint_path = get_last_checkpoint(sft_args.output_dir)
            else:
                resume_from_checkpoint_path = None
            assert (
                resume_from_checkpoint_path is not None
            ), f"cannot find last checkpoint dir in {sft_args.output_dir}"
        elif sft_args.resume_from_checkpoint in [None, "False", "false", False]:
            resume_from_checkpoint_bool = False
            resume_from_checkpoint_path = None
        else:
            resume_from_checkpoint_bool = True
            resume_from_checkpoint_path = os.path.join(
                sft_args.output_dir, sft_args.resume_from_checkpoint
            )
            assert os.path.isdir(
                resume_from_checkpoint_path
            ), f"{resume_from_checkpoint_path} is not a dir."

        if (
            output_dir_contains_file
            and not sft_args.overwrite_output_dir
            and not resume_from_checkpoint_bool
        ):
            raise ValueError(
                f"Output_dir ({sft_args.output_dir}) is not empty. Maybe you mean --resume_from_checkpoint"
                '="True" to resume a training or --overwrite_output_dir to overwrite output_dir.'
            )

        return resume_from_checkpoint_path

    def map_fn(self, dataset, **kwargs):
        """
        dataset map and template function. The default implement follows the BelleGroup/train_0.5M_CN format, means
        'instruction', 'input' and 'output' are necessary. Since some other popular dataset like tatsu-lab/alpaca
        provides these columns as well, it is also supported.
        """
        args: HFSftArgs = kwargs.get("args", None)
        instruction = dataset["instruction"]
        input_text = dataset["input"]
        output_text = dataset["output"]
        bos_token = self.tokenizer.bos_token or ""
        eos_token = self.tokenizer.eos_token
        input_prompt = f"{bos_token}{instruction} {input_text}{eos_token}"
        tokenized_full_prompt = self._tokenize_dataset(
            input_prompt, args.max_input_length
        )
        return tokenized_full_prompt

    def _init_dataset(
        self, args: HFSftArgs
    ) -> typing.Tuple[Union[Dataset], Union[Dataset]]:  # noqa
        """
        init and map dataset, for train and eval
        """
        with args.main_process_first(desc="initialize dataset"):
            train_dataset = None
            if args.train_dataset_path:
                train_dataset = (
                    self._load_dataset(args.train_dataset_path, "train")
                    .shuffle()
                    .map(self.map_fn, fn_kwargs={"args": args})
                )

            eval_dataset = None
            if args.eval_dataset_path:
                eval_dataset = (
                    self._load_dataset(args.eval_dataset_path, "train")
                    .shuffle()
                    .map(self.map_fn, fn_kwargs={"args": args})
                )

            return train_dataset, eval_dataset

    def _load_dataset(self, data_path, split="train"):  # noqa
        from nn4k.utils.io.dataset_utils import DatasetUtils

        return DatasetUtils.auto_dataset(data_path, split)

    def load_model(self, args: dict = None, mode=None, **kwargs):
        """
        load model and tokenizer. If the model with the same mode is already loaded, will not load again.
        """

        assert (
            mode is not None
        ), f"mode should be either 'train' or 'inference' for HFLLMExecutor, {mode} is illegal."

        if self.model_mode == mode and self._model is not None:
            return

        args = args or self._init_args

        from transformers import HfArgumentParser
        from nn4k.executor.huggingface import HFModelArgs

        parser = HfArgumentParser(HFModelArgs)
        hf_model_args: HFModelArgs
        hf_model_args, *_ = parser.parse_dict(args, allow_extra_keys=True)

        self.model_mode = mode
        self._tokenizer = self._hf_tokenizer_loader(hf_model_args)
        self._model = self._hf_model_loader(
            args=hf_model_args, mode=mode, device=hf_model_args.nn_device
        )

        if self.tokenizer.eos_token_id is None:
            self.tokenizer.eos_token_id = self.model.config.eos_token_id
        if self.tokenizer.pad_token_id is None:
            self.tokenizer.pad_token_id = self.tokenizer.eos_token_id
        if hf_model_args.padding_side is not None:
            self.tokenizer.padding_side = hf_model_args.padding_side

    def inference(self, inputs, **kwargs):
        infer_args = ArgsUtils.update_args(self.init_args, kwargs)

        from transformers import HfArgumentParser

        parser = HfArgumentParser(HFInferArgs)
        hf_infer_args: HFInferArgs
        hf_infer_args, *_ = parser.parse_dict(infer_args, allow_extra_keys=True)

        model = self.model
        tokenizer = self.tokenizer

        input_ids = tokenizer(
            inputs,
            **hf_infer_args.tokenize_config,
        ).to(model.device)

        if hf_infer_args.stop_sequence is not None:
            stop_sequence = hf_infer_args.stop_sequence
            stop_sequence_ids = self.tokenizer.encode(
                stop_sequence, add_special_tokens=False
            )
            if len(stop_sequence_ids) > 1:
                print(  # TODO: use logger instead
                    "Warning: Stopping on a multiple token sequence is not yet supported on transformers. "
                    "The first token of the stop sequence will be used as the stop sequence string in the interim."
                )
            hf_infer_args.generate_config["eos_token_id"] = stop_sequence_ids[0]

        output_ids = model.generate(
            **input_ids,
            **hf_infer_args.generate_config,
        )

        output_texts = []
        for idx, output_id in enumerate(output_ids):
            if not hf_infer_args.return_input_text:
                output_id = output_id[len(input_ids["input_ids"][idx]) :]
            output_text = self.tokenizer.decode(
                output_id, **hf_infer_args.decode_config
            )

            if (
                not hf_infer_args.return_input_text
                and hf_infer_args.delete_heading_new_lines
            ):
                import re

                match = re.search("(\\n)+", output_text)
                if match is not None:
                    start_index = match.end()
                    if start_index < len(output_text) - 1:
                        output_text = output_text[start_index:]

            output_texts.append(output_text)

        return output_texts

    @abstractmethod
    def _hf_model_loader(
        self,
        args: HFModelArgs,
        mode,
        resume_from_checkpoint=False,
        device=None,
        **kwargs,
    ):
        """
        load model into given device for hugging face.
        """
        pass

    def _hf_tokenizer_loader(self, args: HFModelArgs, **kwargs):  # noqa
        """
        hugging face tokenizer loader
        """
        tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=args.pretrained_model_name_or_path,
            use_fast=False,
            revision=args.nn_version,
            trust_remote_code=args.trust_remote_code,
        )
        return tokenizer

    def _hf_model_config_loader(self, args: HFModelArgs, **kwargs):  # noqa
        """
        hugging face model config loader
        """
        model_config = AutoConfig.from_pretrained(
            args.pretrained_model_name_or_path,
            trust_remote_code=args.trust_remote_code,
            **kwargs,
        )
        return model_config

    def _init_trainer(
        self, train_dataset, eval_dataset, sft_args: HFSftArgs, callbacks=None
    ) -> Trainer:
        """
        hugging face model trainer initializer
        """
        trainer = NNHFTrainer(
            model=self.model,
            args=sft_args,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=self.tokenizer,
            data_collator=self._data_collator(),
            callbacks=callbacks,
        )

        return trainer

    @abstractmethod
    def _data_collator(self, return_tensors="pt", **kwargs):
        """
        data collator used in trainer
        """
        pass

    def _tokenize_dataset(self, prompt_text, max_length):
        """
        tokenize dataset, by default will cut the input to the max_length
        """
        tokenized_dataset = self.tokenizer(
            prompt_text, truncation=True, max_length=max_length
        )
        input_ids = tokenized_dataset["input_ids"]
        attention_mask = tokenized_dataset["attention_mask"]

        # append eos token if necessary
        # input length is shorter than max_length
        if len(input_ids) < max_length:
            if input_ids[-1] != self.tokenizer.eos_token_id:
                input_ids.append(self.tokenizer.eos_token_id)
                attention_mask.append(1)
        else:
            input_ids[max_length - 1] = self.tokenizer.eos_token_id
            attention_mask[max_length - 1] = 1

        # labels are copy of input_ids
        tokenized_dataset["labels"] = tokenized_dataset["input_ids"].copy()

        return tokenized_dataset



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


class ArgsUtils:
    CONFIG_FILE_KEY = "config_file"

    @staticmethod
    def update_args(base_args: dict, new_args: dict) -> dict:
        """
        update an existing args with a new set of args
        :param base_args: args to get updated. Will be copied before get updated.
        :param new_args: args to update the base args.
        :rtype: dict
        """
        import copy

        copy_base_args = copy.deepcopy(base_args)
        new_args = new_args or {}
        copy_base_args.update(new_args)
        return copy_base_args

    @staticmethod
    def handle_dict_config(kwargs: dict) -> dict:
        if "config_file" in kwargs:
            configs = ArgsUtils.load_config_dict_from_file(kwargs.get("config_file"))
        else:
            configs = kwargs

        return configs

    @staticmethod
    def load_config_dict_from_file(file_path: str) -> dict:
        from pathlib import Path

        if file_path.endswith(".json"):
            import json

            with open(Path(file_path), "r", encoding="utf-8") as open_json_file:
                data = json.load(open_json_file)
                nn_config = data
                return nn_config
        if file_path.endswith(".json5"):
            import json5

            with open(Path(file_path), "r", encoding="utf-8") as open_json5_file:
                data = json5.load(open_json5_file)
                nn_config = data
                return nn_config
        from nn4k.utils.io.file_utils import FileUtils

        raise ValueError(
            f"Config file with extension type {FileUtils.get_extension(file_path)} is not supported."
            f"use json or json5 instead."
        )
